@@ -5,8 +5,10 @@ import enum
 from typing import Mapping, Optional, Sequence, Union
 
 from .expr import (
+    BaseExpr,
     EnableExpr,
     IdExpr,
+    MappingT,
     OptBoolExpr,
     OptIdExpr,
     OptIntExpr,
@@ -14,8 +16,8 @@ from .expr import (
     OptRemotePathExpr,
     OptStrExpr,
     OptTimeDeltaExpr,
-    PortPairExpr,
     RemotePathExpr,
+    SequenceT,
     SimpleIdExpr,
     SimpleOptBoolExpr,
     SimpleOptIdExpr,
@@ -75,9 +77,10 @@ class Image(Base):
     ref: StrExpr  # Image reference, e.g. image:my-proj or neuromation/base@v1.6
     context: OptLocalPathExpr
     dockerfile: OptLocalPathExpr
-    build_args: Optional[Sequence[StrExpr]] = field(metadata={"allow_none": True})
-    env: Optional[Mapping[str, StrExpr]] = field(metadata={"allow_none": True})
-    volumes: Optional[Sequence[OptStrExpr]] = field(metadata={"allow_none": True})
+    build_args: Optional[BaseExpr[SequenceT]] = field(metadata={"allow_none": True})
+    env: Optional[BaseExpr[MappingT]] = field(metadata={"allow_none": True})
+    volumes: Optional[BaseExpr[SequenceT]] = field(metadata={"allow_none": True})
+    build_preset: OptStrExpr
 
 
 @dataclass(frozen=True)
@@ -90,9 +93,9 @@ class ExecUnit(Base):
     entrypoint: OptStrExpr
     cmd: OptStrExpr
     workdir: OptRemotePathExpr
-    env: Optional[Mapping[str, StrExpr]] = field(metadata={"allow_none": True})
-    volumes: Optional[Sequence[OptStrExpr]] = field(metadata={"allow_none": True})
-    tags: Optional[Sequence[StrExpr]] = field(metadata={"allow_none": True})
+    env: Optional[BaseExpr[MappingT]] = field(metadata={"allow_none": True})
+    volumes: Optional[BaseExpr[SequenceT]] = field(metadata={"allow_none": True})
+    tags: Optional[BaseExpr[SequenceT]] = field(metadata={"allow_none": True})
     life_span: OptTimeDeltaExpr
     http_port: OptIntExpr
     http_auth: OptBoolExpr
@@ -139,12 +142,15 @@ class Job(ExecUnit, JobBase):
 
     detach: OptBoolExpr
     browse: OptBoolExpr
-    port_forward: Optional[Sequence[PortPairExpr]] = field(
-        metadata={"allow_none": True}
-    )
+    port_forward: Optional[BaseExpr[SequenceT]] = field(metadata={"allow_none": True})
     multi: SimpleOptBoolExpr
 
 
+class NeedsLevel(enum.Enum):
+    RUNNING = "running"
+    COMPLETED = "completed"  # pipelined mode
+
+
 @dataclass(frozen=True)
 class TaskBase(Base):
     id: OptIdExpr
@@ -152,7 +158,7 @@ class TaskBase(Base):
     # A set of steps, used in net mode
     # All steps share the same implicit persistent disk volume
 
-    needs: Optional[Sequence[IdExpr]] = field(metadata={"allow_none": True})
+    needs: Optional[Mapping[IdExpr, NeedsLevel]] = field(metadata={"allow_none": True})
 
     # matrix? Do we need a build matrix? Yes probably.
 
@@ -187,9 +193,9 @@ class TaskActionCall(BaseActionCall, TaskBase):
 
 @dataclass(frozen=True)
 class FlowDefaults(Base):
-    tags: Optional[Sequence[StrExpr]] = field(metadata={"allow_none": True})
+    tags: Optional[BaseExpr[SequenceT]] = field(metadata={"allow_none": True})
 
-    env: Optional[Mapping[str, StrExpr]] = field(metadata={"allow_none": True})
+    env: Optional[BaseExpr[MappingT]] = field(metadata={"allow_none": True})
     workdir: OptRemotePathExpr
 
     life_span: OptTimeDeltaExpr
@@ -289,7 +295,6 @@ class BatchAction(BaseAction):
 @dataclass(frozen=True)
 class StatefulAction(BaseAction):
     outputs: Optional[Mapping[str, Output]] = field(metadata={"allow_none": True})
-    cache: Optional[Cache] = field(metadata={"allow_none": True})
     main: ExecUnit
     post: Optional[ExecUnit] = field(metadata={"allow_none": True})
     post_if: EnableExpr = field(metadata={"default_expr": "${{ always() }}"})
