@@ -3,10 +3,9 @@ import dataclasses
 import click
 import datetime
 import neuro_extras
-import neuromation
-import sys
 from graphviz import Digraph
-from neuromation.api import Client, ResourceNotFound
+from neuro_cli import __version__ as cli_version
+from neuro_sdk import Client, ResourceNotFound, __version__ as sdk_version
 from operator import attrgetter, itemgetter
 from rich import box
 from rich.console import Console
@@ -19,6 +18,7 @@ from typing_extensions import AsyncContextManager, AsyncIterator
 import neuro_flow
 
 from .batch_executor import BatchExecutor, ExecutorData, LocalsBatchExecutor
+from .colored_topo_sorter import ColoredTopoSorter
 from .commands import CmdProcessor
 from .config_loader import BatchLocalCL
 from .context import EMPTY_ROOT, EarlyBatch, RunningBatchFlow
@@ -28,11 +28,6 @@ from .types import FullID, LocalPath, TaskStatus
 from .utils import TERMINATED_TASK_STATUSES, fmt_datetime, run_subproc
 
 
-if sys.version_info >= (3, 9):
-    import graphlib
-else:
-    from . import backport_graphlib as graphlib
-
 EXECUTOR_IMAGE = f"neuromation/neuro-flow:{neuro_flow.__version__}"
 
 
@@ -77,7 +72,7 @@ class BatchRunner(AsyncContextManager["BatchRunner"]):
             await self._config_loader.close()
 
     async def __aenter__(self) -> "BatchRunner":
-        self._config_loader = BatchLocalCL(self._config_dir)
+        self._config_loader = BatchLocalCL(self._config_dir, self._client)
         project_ast = await self._config_loader.fetch_project()
         self._project = await project_ast.id.eval(EMPTY_ROOT)
         return self
@@ -96,7 +91,8 @@ class BatchRunner(AsyncContextManager["BatchRunner"]):
     ) -> ExecutorData:
         # batch_name is a name of yaml config inside self._workspace / .neuro
         # folder without the file extension
-        self._console.log(f"[bright_black]neuromation=={neuromation.__version__}")
+        self._console.log(f"[bright_black]neuro_sdk=={sdk_version}")
+        self._console.log(f"[bright_black]neuro_cli=={cli_version}")
         self._console.log(f"[bright_black]neuro-extras=={neuro_extras.__version__}")
         self._console.log(f"[bright_black]neuro-flow=={neuro_flow.__version__}")
         self._console.log(f"Use config file {self.config_loader.flow_path(batch_name)}")
@@ -150,8 +146,7 @@ class BatchRunner(AsyncContextManager["BatchRunner"]):
             }
 
             # check fast for the graph cycle error
-            toposorter = graphlib.TopologicalSorter(graph)
-            toposorter.prepare()
+            ColoredTopoSorter(flow.graph)
 
             for tid in graph:
                 if await flow.is_action(tid):
@@ -198,9 +193,7 @@ class BatchRunner(AsyncContextManager["BatchRunner"]):
             await run_subproc(
                 "neuro",
                 "run",
-                # TODO: re-enable restarting following issue is fixed:
-                # https://github.com/neuro-inc/platform-client-python/issues/1797
-                # "--restart=on-failure",
+                "--restart=on-failure",
                 "--pass-config",
                 EXECUTOR_IMAGE,
                 "neuro-flow",
@@ -489,6 +482,14 @@ class BatchRunner(AsyncContextManager["BatchRunner"]):
                     graph = graphs[prefix]
                     deps = graph[task.id]
                     if not deps or all(dep in handled for dep in deps):
+                        if (
+                            prefix in finished
+                            and finished[prefix].status != TaskStatus.SUCCEEDED
+                        ):
+                            # If action did not succeeded, we should write started
+                            await self._storage.write_start(
+                                dataclasses.replace(started[prefix], attempt=new_att)
+                            )
                         await self._storage.write_start(
                             dataclasses.replace(started[task.id], attempt=new_att)
                         )
