@@ -7,11 +7,14 @@ import hashlib
 import itertools
 import json
 import shlex
+import sys
 from abc import abstractmethod
 from functools import lru_cache
+from neuro_sdk import Client
 from typing import (
     AbstractSet,
     Any,
+    AsyncIterator,
     Dict,
     Generic,
     Iterable,
@@ -33,6 +36,11 @@ from neuro_flow.expr import EnableExpr, EvalError, LiteralT, RootABC, TypeT
 from neuro_flow.types import AlwaysT, FullID, LocalPath, RemotePath, TaskStatus
 
 
+if sys.version_info >= (3, 7):  # pragma: no cover
+    from contextlib import asynccontextmanager
+else:
+    from async_generator import asynccontextmanager
+
 # Exceptions
 
 
@@ -117,6 +125,7 @@ class ImageCtx:
     build_args: Sequence[str]
     env: Mapping[str, str]
     volumes: Sequence[str]
+    build_preset: Optional[str]
 
 
 @dataclass(frozen=True)
@@ -233,12 +242,19 @@ class EmptyRoot(RootABC):
     def lookup(self, name: str) -> TypeT:
         raise NotAvailable(name)
 
+    @asynccontextmanager
+    async def client(self) -> AsyncIterator[Client]:
+        raise RuntimeError("neuro API is not available in <empty> context")
+        yield Client()  # fake lint to make the code a real async iterator
+
 
 EMPTY_ROOT = EmptyRoot()
 
 
 @dataclass(frozen=True)
 class Context(RootABC):
+    _client: Client
+
     def lookup(self, name: str) -> TypeT:
         for f in fields(self):
             if f.name != name:
@@ -250,6 +266,10 @@ class Context(RootABC):
         # assert isinstance(ret, (ContainerT, SequenceT, MappingT)), ret
         return cast(TypeT, ret)
 
+    @asynccontextmanager
+    async def client(self) -> AsyncIterator[Client]:
+        yield self._client
+
 
 @dataclass(frozen=True)
 class WithFlowContext(Context):
@@ -272,6 +292,7 @@ class LiveContextStep1(WithFlowContext, Context):
             tags=tags,
             volumes=volumes,
             images=images,
+            _client=self._client,
         )
 
 
@@ -289,6 +310,7 @@ class LiveContext(WithEnvContext, LiveContextStep1):
             volumes=self.volumes,
             images=self.images,
             params=params,
+            _client=self._client,
         )
 
     def to_multi_job_ctx(
@@ -302,6 +324,7 @@ class LiveContext(WithEnvContext, LiveContextStep1):
             images=self.images,
             multi=multi,
             params=params,
+            _client=self._client,
         )
 
 
@@ -317,8 +340,7 @@ class LiveMultiJobContext(LiveContext):
 
 
 @dataclass(frozen=True)
-class LiveActionContext(WithEnvContext, Context):
-    tags: TagsCtx
+class LiveActionContext(Context):
     inputs: InputsCtx
 
 
@@ -336,6 +358,7 @@ class BatchContextStep1(WithFlowContext, Context):
             tags=tags,
             volumes=volumes,
             images=images,
+            _client=self._client,
         )
 
 
@@ -357,11 +380,11 @@ class BatchContextStep2(WithEnvContext, BatchContextStep1):
             volumes=self.volumes,
             images=self.images,
             strategy=strategy,
+            _client=self._client,
         )
 
 
-class BaseBatchContext(WithEnvContext, Context):
-    tags: TagsCtx
+class BaseBatchContext(Context):
     strategy: StrategyCtx
 
     @abstractmethod
@@ -406,6 +429,7 @@ class BatchContext(BaseBatchContext, BatchContextStep2):
             images=self.images,
             strategy=strategy,
             matrix=matrix,
+            _client=self._client,
         )
 
 
@@ -425,6 +449,7 @@ class BatchMatrixContext(BaseMatrixContext, BatchContext):
             matrix=self.matrix,
             needs=needs,
             state=state,
+            _client=self._client,
         )
 
 
@@ -436,29 +461,25 @@ class BatchTaskContext(BaseTaskContext, BatchMatrixContext):
 
 @dataclass(frozen=True)
 class BatchActionContext(BaseBatchContext):
-    env: EnvCtx
     inputs: InputsCtx
-    tags: TagsCtx
     strategy: StrategyCtx
 
     def to_matrix_ctx(
         self, strategy: StrategyCtx, matrix: MatrixCtx
     ) -> "BatchActionMatrixContext":
         return BatchActionMatrixContext(
-            env=self.env,
-            tags=self.tags,
             inputs=self.inputs,
             matrix=matrix,
             strategy=strategy,
+            _client=self._client,
         )
 
     def to_outputs_ctx(self, needs: NeedsCtx) -> "BatchActionOutputsContext":
         return BatchActionOutputsContext(
-            env=self.env,
-            tags=self.tags,
             strategy=self.strategy,
             inputs=self.inputs,
             needs=needs,
+            _client=self._client,
         )
 
 
@@ -474,13 +495,12 @@ class BatchActionMatrixContext(BaseMatrixContext, BatchActionContext):
 
     def to_task_ctx(self, needs: NeedsCtx, state: StateCtx) -> "BatchActionTaskContext":
         return BatchActionTaskContext(
-            env=self.env,
-            tags=self.tags,
             inputs=self.inputs,
             matrix=self.matrix,
             strategy=self.strategy,
             needs=needs,
             state=state,
+            _client=self._client,
         )
 
 
@@ -491,14 +511,12 @@ class BatchActionTaskContext(BaseTaskContext, BatchActionMatrixContext):
 
 
 @dataclass(frozen=True)
-class StatefulActionContext(WithEnvContext, Context):
-    tags: TagsCtx
+class StatefulActionContext(Context):
     inputs: InputsCtx
 
 
 @dataclass(frozen=True)
-class LocalActionContext(WithEnvContext, Context):
-    tags: TagsCtx
+class LocalActionContext(Context):
     inputs: InputsCtx
 
 
@@ -528,14 +546,20 @@ async def setup_defaults_env_tags_ctx(
     ctx: WithFlowContext,
     ast_defaults: Optional[ast.FlowDefaults],
 ) -> Tuple[DefaultsConf, EnvCtx, TagsCtx]:
+    env: EnvCtx
+    tags: TagsCtx
     if ast_defaults is not None:
         if ast_defaults.env is not None:
-            env = {k: await v.eval(ctx) for k, v in ast_defaults.env.items()}
+            tmp_env = await ast_defaults.env.eval(ctx)
+            assert isinstance(tmp_env, dict)
+            env = tmp_env
         else:
             env = {}
 
         if ast_defaults.tags is not None:
-            tags = {await t.eval(ctx) for t in ast_defaults.tags}
+            tmp_tags = await ast_defaults.tags.eval(ctx)
+            assert isinstance(tmp_tags, list)
+            tags = set(tmp_tags)
         else:
             tags = set()
         workdir = await ast_defaults.workdir.eval(ctx)
@@ -600,21 +624,25 @@ async def setup_images_ctx(
         for k, i in ast_images.items():
             context_path = await i.context.eval(ctx)
             dockerfile_path = await i.dockerfile.eval(ctx)
+            build_args: List[str] = []
             if i.build_args is not None:
-                build_args = [await v.eval(ctx) for v in i.build_args]
-            else:
-                build_args = []
+                tmp_build_args = await i.build_args.eval(ctx)
+                assert isinstance(tmp_build_args, list)
+                build_args = tmp_build_args
 
-            image_env = {}
+            image_env: Dict[str, str] = {}
             if i.env is not None:
-                image_env.update({k: await v.eval(ctx) for k, v in i.env.items()})
+                tmp_env = await i.env.eval(ctx)
+                assert isinstance(tmp_env, dict)
+                image_env.update(tmp_env)
 
-            image_volumes = []
+            image_volumes: List[str] = []
             if i.volumes is not None:
-                for vol in i.volumes:
-                    value = await vol.eval(ctx)
-                    if value:
-                        image_volumes.append(value)
+                tmp_volumes = await i.volumes.eval(ctx)
+                assert isinstance(tmp_volumes, list)
+                for volume in tmp_volumes:
+                    if volume:
+                        image_volumes.append(volume)
 
             images[k] = ImageCtx(
                 id=k,
@@ -626,6 +654,7 @@ async def setup_images_ctx(
                 build_args=build_args,
                 env=image_env,
                 volumes=image_volumes,
+                build_preset=await i.build_preset.eval(ctx),
             )
     return images
 
@@ -751,6 +780,22 @@ async def setup_matrix(ast_matrix: Optional[ast.Matrix]) -> Sequence[MatrixCtx]:
     matrices = filtered
     # Include
     for inc_spec in ast_matrix.include:
+        if inc_spec.keys() != ast_matrix.products.keys():
+            additional = inc_spec.keys() - ast_matrix.products.keys()
+            missing = ast_matrix.products.keys() - inc_spec.keys()
+            raise EvalError(
+                "Keys of entry in include list of matrix are not the "
+                "same as matrix keys: "
+                + (
+                    f"additional keys: {','.join(sorted(additional))}"
+                    if additional
+                    else ""
+                )
+                + (f" , " if additional and missing else "")
+                + (f"missing keys: {','.join(sorted(missing))}" if missing else ""),
+                ast_matrix._start,
+                ast_matrix._end,
+            )
         matrices.append({k: await v.eval(EMPTY_ROOT) for k, v in inc_spec.items()})
     return matrices
 
@@ -878,7 +923,7 @@ class RunningLiveFlow:
         ctx = self._ctx.to_job_ctx(
             params=await setup_params_ctx(EMPTY_ROOT, params, job_ast.params)
         )
-        return await self._get_job(ctx, self._defaults, job_id, params)
+        return await self._get_job(ctx, ctx.env, self._defaults, job_id)
 
     async def get_multi_job(
         self,
@@ -900,35 +945,39 @@ class RunningLiveFlow:
             multi=MultiCtx(suffix=suffix, args=args_str),
             params=await setup_params_ctx(EMPTY_ROOT, params, job_ast.params),
         )
-        job = await self._get_job(ctx, self._defaults, job_id, params)
+        job = await self._get_job(ctx, ctx.env, self._defaults, job_id)
         return replace(job, tags=job.tags | {f"multi:{suffix}"})
 
     async def _get_job(
         self,
-        ctx: WithEnvContext,
+        ctx: RootABC,
+        env_ctx: EnvCtx,
         defaults: DefaultsConf,
         job_id: str,
-        params: Mapping[str, str],
     ) -> Job:
         job = self._get_job_ast(job_id)
         if isinstance(job, ast.JobActionCall):
             action_ast = await self._get_action_ast(job)
             ctx = LiveActionContext(
-                tags=self._ctx.tags,
-                env={},  # TODO: Is it correct?
                 inputs=await setup_inputs_ctx(ctx, job, action_ast.inputs),
+                _client=self._ctx._client,
             )
+            env_ctx = {}
             defaults = DefaultsConf()
             job = action_ast.job
         assert isinstance(job, ast.Job)
 
         tags = (await self.get_meta(job_id)).tags
         if job.tags is not None:
-            tags |= {await v.eval(ctx) for v in job.tags}
+            tmp_tags = await job.tags.eval(ctx)
+            assert isinstance(tmp_tags, list)
+            tags |= set(tmp_tags)
 
-        env = dict(ctx.env)
+        env = dict(env_ctx)
         if job.env is not None:
-            env.update({k: await v.eval(ctx) for k, v in job.env.items()})
+            tmp_env = await job.env.eval(ctx)
+            assert isinstance(tmp_env, dict)
+            env.update(tmp_env)
 
         title = await job.title.eval(ctx)
         if title is None:
@@ -936,12 +985,13 @@ class RunningLiveFlow:
 
         workdir = (await job.workdir.eval(ctx)) or defaults.workdir
 
-        volumes = []
+        volumes: List[str] = []
         if job.volumes is not None:
-            for v in job.volumes:
-                val = await v.eval(ctx)
-                if val:
-                    volumes.append(val)
+            tmp_volumes = await job.volumes.eval(ctx)
+            assert isinstance(tmp_volumes, list)
+            for volume in tmp_volumes:
+                if volume:
+                    volumes.append(volume)
 
         life_span = (await job.life_span.eval(ctx)) or defaults.life_span
 
@@ -949,9 +999,11 @@ class RunningLiveFlow:
         schedule_timeout = (
             await job.schedule_timeout.eval(ctx)
         ) or defaults.schedule_timeout
-        port_forward = []
+        port_forward: List[str] = []
         if job.port_forward is not None:
-            port_forward = [await val.eval(ctx) for val in job.port_forward]
+            tmp_port_forward = await job.port_forward.eval(ctx)
+            assert isinstance(tmp_port_forward, list)
+            port_forward = tmp_port_forward
 
         return Job(
             id=job_id,
@@ -988,7 +1040,7 @@ class RunningLiveFlow:
             EMPTY_ROOT, ast_flow, config_name, config_loader
         )
 
-        step_1_ctx = LiveContextStep1(flow=flow_ctx)
+        step_1_ctx = LiveContextStep1(flow=flow_ctx, _client=config_loader.client)
 
         defaults, env, tags = await setup_defaults_env_tags_ctx(
             step_1_ctx, ast_flow.defaults
@@ -1015,11 +1067,11 @@ class EarlyBatch:
         self._tasks = tasks
 
     @property
-    def graph(self) -> Mapping[str, AbstractSet[str]]:
+    def graph(self) -> Mapping[str, Mapping[str, ast.NeedsLevel]]:
         return self._graph()
 
     @lru_cache()
-    def _graph(self) -> Mapping[str, AbstractSet[str]]:
+    def _graph(self) -> Mapping[str, Mapping[str, ast.NeedsLevel]]:
         # This function is only needed for mypy
         return {key: early_task.needs for key, early_task in self._tasks.items()}
 
@@ -1081,12 +1133,14 @@ class RunningBatchBase(Generic[_T], EarlyBatch):
     def __init__(
         self,
         ctx: _T,
+        default_tags: TagsCtx,
         tasks: Mapping[str, "BasePrepTask"],
         config_loader: ConfigLoader,
         defaults: DefaultsConf,
     ):
         super().__init__(tasks, config_loader)
         self._ctx = ctx
+        self._default_tags = default_tags
         self._defaults = defaults
 
     def _get_prep(self, real_id: str) -> "BasePrepTask":
@@ -1098,9 +1152,14 @@ class RunningBatchBase(Generic[_T], EarlyBatch):
         self, real_id: str, needs: NeedsCtx, state: StateCtx
     ) -> BaseTaskContext:
         prep_task = self._get_prep(real_id)
-        if needs.keys() != prep_task.needs:
-            extra = ",".join(needs.keys() - prep_task.needs)
-            missing = ",".join(prep_task.needs - needs.keys())
+        needs_completed = {
+            task_id
+            for task_id, level in prep_task.needs.items()
+            if level == ast.NeedsLevel.COMPLETED
+        }
+        if needs.keys() != needs_completed:
+            extra = ",".join(needs.keys() - needs_completed)
+            missing = ",".join(needs_completed - needs.keys())
             err = ["Error in 'needs':"]
             if extra:
                 err.append(f"unexpected keys {extra}")
@@ -1136,42 +1195,48 @@ class RunningBatchBase(Generic[_T], EarlyBatch):
         assert isinstance(prep_task, (PrepTask, PrepStatefulCall))  # Already checked
 
         task_ctx = self._task_context(real_id, needs, state)
-        ctx: WithEnvContext = task_ctx
+        ctx: RootABC = task_ctx
         defaults = self._defaults
 
         if isinstance(prep_task, PrepStatefulCall):
             ctx = StatefulActionContext(
-                tags=self._ctx.tags,
-                env={},  # TODO: Is it correct?
                 inputs=await setup_inputs_ctx(
                     ctx, prep_task.call, prep_task.action.inputs
                 ),
+                _client=self._ctx._client,
             )
             defaults = DefaultsConf()  # TODO: Is it correct?
 
         full_id = prefix + (real_id,)
 
-        env = dict(ctx.env)
+        if isinstance(ctx, BatchTaskContext):
+            env = dict(ctx.env)
+        else:
+            env = {}
+
         if prep_task.ast_task.env is not None:
-            env.update(
-                {k: await v.eval(ctx) for k, v in prep_task.ast_task.env.items()}
-            )
+            tmp_env = await prep_task.ast_task.env.eval(ctx)
+            assert isinstance(tmp_env, dict)
+            env.update(tmp_env)
 
         title = await prep_task.ast_task.title.eval(ctx)
 
         tags = set()
         if prep_task.ast_task.tags is not None:
-            tags = {await v.eval(ctx) for v in prep_task.ast_task.tags}
-        if not tags:
-            tags = {"task:" + _id2tag(".".join(full_id))}
-        tags |= set(self._ctx.tags)
+            tmp_tags = await prep_task.ast_task.tags.eval(ctx)
+            assert isinstance(tmp_tags, list)
+            tags |= set(tmp_tags)
+
+        tags |= {"task:" + _id2tag(".".join(full_id))}
+        tags |= set(self._default_tags)
 
         workdir = (await prep_task.ast_task.workdir.eval(ctx)) or defaults.workdir
 
-        volumes = []
+        volumes: List[str] = []
         if prep_task.ast_task.volumes is not None:
-            for v in prep_task.ast_task.volumes:
-                val = await v.eval(ctx)
+            tmp_volumes = await prep_task.ast_task.volumes.eval(ctx)
+            assert isinstance(tmp_volumes, list)
+            for val in tmp_volumes:
                 if val:
                     volumes.append(val)
 
@@ -1224,7 +1289,7 @@ class RunningBatchBase(Generic[_T], EarlyBatch):
             base_cache=prep_task.cache,
             base_strategy=prep_task.strategy,
             inputs=await setup_inputs_ctx(ctx, prep_task.call, prep_task.action.inputs),
-            tags=ctx.tags,
+            default_tags=self._default_tags,
             config_loader=self._cl,
         )
 
@@ -1237,7 +1302,7 @@ class RunningBatchFlow(RunningBatchBase[BatchContext]):
         config_loader: ConfigLoader,
         defaults: DefaultsConf,
     ):
-        super().__init__(ctx, tasks, config_loader, defaults)
+        super().__init__(ctx, ctx.tags, tasks, config_loader, defaults)
 
     @property
     def project_id(self) -> str:
@@ -1261,9 +1326,8 @@ class RunningBatchFlow(RunningBatchBase[BatchContext]):
         ctx = self._task_context(real_id, needs, {})
 
         action_ctx = LocalActionContext(
-            tags=self._ctx.tags,  # TODO: do we need tags for local actions?
-            env={},  # TODO: Is it correct?
             inputs=await setup_inputs_ctx(ctx, prep_task.call, prep_task.action.inputs),
+            _client=self._ctx._client,
         )
 
         return LocalTask(
@@ -1288,6 +1352,7 @@ class RunningBatchFlow(RunningBatchBase[BatchContext]):
         step_1_ctx = BatchContextStep1(
             flow=flow_ctx,
             params=params_ctx,
+            _client=config_loader.client,
         )
 
         defaults, env, tags = await setup_defaults_env_tags_ctx(
@@ -1324,13 +1389,14 @@ class RunningBatchActionFlow(RunningBatchBase[BatchActionContext]):
     def __init__(
         self,
         ctx: BatchActionContext,
+        default_tags: TagsCtx,
         tasks: Mapping[str, "BasePrepTask"],
         config_loader: ConfigLoader,
         defaults: DefaultsConf,
         action: ast.BatchAction,
         output_needs: AbstractSet[str],
     ):
-        super().__init__(ctx, tasks, config_loader, defaults)
+        super().__init__(ctx, default_tags, tasks, config_loader, defaults)
         self._action = action
         self._output_needs = output_needs
 
@@ -1370,14 +1436,13 @@ class RunningBatchActionFlow(RunningBatchBase[BatchActionContext]):
         base_cache: CacheConf,
         base_strategy: StrategyCtx,
         inputs: InputsCtx,
-        tags: TagsCtx,
+        default_tags: TagsCtx,
         config_loader: ConfigLoader,
     ) -> "RunningBatchActionFlow":
         action_context = BatchActionContext(
-            tags=tags,
-            env={},
             inputs=inputs,
             strategy=base_strategy,
+            _client=config_loader.client,
         )
 
         cache = await setup_cache(
@@ -1392,6 +1457,7 @@ class RunningBatchActionFlow(RunningBatchBase[BatchActionContext]):
 
         return RunningBatchActionFlow(
             action_context,
+            default_tags,
             tasks,
             config_loader,
             DefaultsConf(),
@@ -1407,7 +1473,7 @@ class RunningBatchActionFlow(RunningBatchBase[BatchActionContext]):
 class BaseEarlyTask:
     id: Optional[str]
     real_id: str
-    needs: AbstractSet[str]  # A set of batch.id
+    needs: Mapping[str, ast.NeedsLevel]  # Keys are batch.id
 
     matrix: MatrixCtx
     enable: EnableExpr
@@ -1564,7 +1630,7 @@ class BasePrepTask(BaseEarlyTask):
             needs=self.needs,
             matrix=self.matrix,
             strategy=self.strategy,
-            cache=self.cache,
+            cache=CacheConf(strategy=ast.CacheStrategy.NONE),
             enable=self.enable,
             action=action,
             call=call,
@@ -1581,7 +1647,7 @@ class BasePrepTask(BaseEarlyTask):
             needs=self.needs,
             matrix=self.matrix,
             strategy=self.strategy,
-            cache=self.cache,
+            cache=CacheConf(strategy=ast.CacheStrategy.NONE),
             enable=self.enable,
             ast_task=action.main,
             action=action,
@@ -1595,7 +1661,7 @@ class BasePrepTask(BaseEarlyTask):
             needs=self.needs,
             matrix=self.matrix,
             strategy=self.strategy,
-            cache=self.cache,
+            cache=CacheConf(strategy=ast.CacheStrategy.NONE),
             enable=self.enable,
             ast_task=ast_task,
             state_from=state_from,
@@ -1665,7 +1731,10 @@ class EarlyTaskGraphBuilder:
             post_tasks_group = []
             for matrix in matrices:
                 # make prep patch(es)
-                matrix_ctx = MatrixOnlyContext(matrix=matrix)
+                matrix_ctx = MatrixOnlyContext(
+                    matrix=matrix,
+                    _client=self._cl.client,
+                )
 
                 task_id, real_id = await self._setup_ids(matrix_ctx, num, ast_task)
                 needs = await self._setup_needs(matrix_ctx, last_needs, ast_task)
@@ -1685,6 +1754,14 @@ class EarlyTaskGraphBuilder:
                     assert isinstance(ast_task, ast.TaskActionCall)
                     action_name = await ast_task.action.eval(EMPTY_ROOT)
                     action = await self._cl.fetch_action(action_name)
+                    if ast_task.cache and not isinstance(action, ast.BatchAction):
+                        raise EvalError(
+                            f"Specifying cache in action call to the action "
+                            f"{action_name} of kind {action.kind.value} is "
+                            f"not supported.",
+                            ast_task._start,
+                            ast_task._end,
+                        )
                     if isinstance(action, ast.BatchAction):
                         prep_tasks[real_id] = base.to_batch_call(action, ast_task)
                     elif isinstance(action, ast.LocalAction):
@@ -1696,7 +1773,7 @@ class EarlyTaskGraphBuilder:
                                     base,
                                     id=None,
                                     real_id=f"post-{base.real_id}",
-                                    needs={real_id},
+                                    needs={real_id: ast.NeedsLevel.COMPLETED},
                                     enable=action.post_if,
                                 ).to_post_task(action.post, real_id),
                             )
@@ -1704,7 +1781,7 @@ class EarlyTaskGraphBuilder:
 
                     else:
                         raise ValueError(
-                            f"Action {action_name} has kind {action.kind}, "
+                            f"Action {action_name} has kind {action.kind.value}, "
                             "that is not supported in batch mode."
                         )
                 real_ids.add(real_id)
@@ -1717,7 +1794,9 @@ class EarlyTaskGraphBuilder:
         for post_tasks_group in reversed(post_tasks):
             real_ids = set()
             for task in post_tasks_group:
-                task = replace(task, needs=last_needs | task.needs)
+                needs = {need: ast.NeedsLevel.COMPLETED for need in last_needs}
+                needs = {**needs, **task.needs}
+                task = replace(task, needs=needs)
                 prep_tasks[task.real_id] = task
                 real_ids.add(task.real_id)
             last_needs = real_ids
@@ -1739,10 +1818,12 @@ class EarlyTaskGraphBuilder:
 
     async def _setup_needs(
         self, ctx: RootABC, default_needs: AbstractSet[str], ast_task: ast.TaskBase
-    ) -> AbstractSet[str]:
+    ) -> Mapping[str, ast.NeedsLevel]:
         if ast_task.needs is not None:
-            return {await need.eval(ctx) for need in ast_task.needs}
-        return default_needs
+            return {
+                await need.eval(ctx): level for need, level in ast_task.needs.items()
+            }
+        return {need: ast.NeedsLevel.COMPLETED for need in default_needs}
 
 
 class TaskGraphBuilder(EarlyTaskGraphBuilder):
@@ -1816,7 +1897,11 @@ def _hash(val: Any) -> str:
 
 def _ctx_default(val: Any) -> Any:
     if dataclasses.is_dataclass(val):
-        return dataclasses.asdict(val)
+        if hasattr(val, "_client"):
+            val = dataclasses.replace(val, _client=None)
+        ret = dataclasses.asdict(val)
+        ret.pop("_client", None)
+        return ret
     elif isinstance(val, enum.Enum):
         return val.value
     elif isinstance(val, RemotePath):
