@@ -33,11 +33,15 @@ from yaml.resolver import Resolver as BaseResolver
 from yaml.scanner import Scanner
 
 from . import ast
-from .ast import BatchActionOutputs
+from .ast import BatchActionOutputs, NeedsLevel
 from .expr import (
+    BaseExpr,
     EnableExpr,
     Expr,
     IdExpr,
+    MappingExpr,
+    MappingItemsExpr,
+    MappingT,
     OptBashExpr,
     OptBoolExpr,
     OptIdExpr,
@@ -49,20 +53,25 @@ from .expr import (
     OptTimeDeltaExpr,
     PortPairExpr,
     RemotePathExpr,
+    SequenceExpr,
+    SequenceItemsExpr,
+    SequenceT,
     SimpleIdExpr,
     SimpleOptBoolExpr,
     SimpleOptIdExpr,
     SimpleOptStrExpr,
     SimpleStrExpr,
     StrExpr,
+    TypeT,
     URIExpr,
+    port_pair_item,
 )
 from .tokenizer import Pos
 from .types import LocalPath
 
 
-_T = TypeVar("_T")
-_Cont = TypeVar("_Cont")
+_T = TypeVar("_T", bound=TypeT)
+_Container = TypeVar("_Container")
 
 
 @dataclasses.dataclass
@@ -103,12 +112,12 @@ def mark2pos(mark: yaml.Mark) -> Pos:
     return Pos(mark.line, mark.column, LocalPath(mark.name))
 
 
-class SimpleCompound(Generic[_T, _Cont], abc.ABC):
+class SimpleCompound(Generic[_T, _Container], abc.ABC):
     def __init__(self, factory: Type[Expr[_T]]) -> None:
         self._factory = factory
 
     @abc.abstractmethod
-    def construct(self, ctor: BaseConstructor, node: yaml.Node) -> _Cont:
+    def construct(self, ctor: BaseConstructor, node: yaml.Node) -> _Container:
         pass
 
     def check_scalar(self, ctor: BaseConstructor, node: yaml.Node) -> None:
@@ -194,6 +203,58 @@ class IdMapping(SimpleCompound[_T, Mapping[str, Expr[_T]]]):
         return ret
 
 
+class ExprOrSeq(SimpleCompound[_T, BaseExpr[SequenceT]]):
+    def __init__(
+        self,
+        item_expr_factory: Type[Expr[_T]],
+        item_value_factory: Callable[[TypeT], TypeT],
+    ) -> None:
+        super().__init__(item_expr_factory)
+        self._item_factory = item_value_factory
+
+    def construct(self, ctor: BaseConstructor, node: yaml.Node) -> BaseExpr[SequenceT]:
+        if isinstance(node, yaml.ScalarNode):
+            val = ctor.construct_object(node)  # type: ignore[no-untyped-call]
+            return SequenceExpr(
+                mark2pos(node.start_mark),
+                mark2pos(node.end_mark),
+                val,
+                self._item_factory,
+            )
+        else:
+            seq = SimpleSeq(self._factory).construct(ctor, node)
+            return SequenceItemsExpr(seq)
+
+
+class ExprOrMapping(SimpleCompound[_T, BaseExpr[MappingT]]):
+    def __init__(
+        self,
+        item_expr_factory: Type[Expr[_T]],
+        item_value_factory: Callable[[TypeT], TypeT],
+    ) -> None:
+        super().__init__(item_expr_factory)
+        self._item_factory = item_value_factory
+
+    def construct(self, ctor: BaseConstructor, node: yaml.Node) -> BaseExpr[MappingT]:
+        if isinstance(node, yaml.ScalarNode):
+            val = ctor.construct_object(node)  # type: ignore[no-untyped-call]
+            return MappingExpr(
+                mark2pos(node.start_mark),
+                mark2pos(node.end_mark),
+                val,
+                self._item_factory,
+            )
+        else:
+            seq = SimpleMapping(self._factory).construct(ctor, node)
+            return MappingItemsExpr(seq)
+
+
+def type2str(arg: TypeT) -> TypeT:
+    if arg is None:
+        return arg
+    return str(arg)
+
+
 _AstType = TypeVar("_AstType", bound=ast.Base)
 _CtorType = TypeVar("_CtorType", bound=BaseConstructor)
 
@@ -474,9 +535,10 @@ def parse_image(ctor: BaseConstructor, node: yaml.MappingNode) -> ast.Image:
             "ref": StrExpr,
             "context": OptLocalPathExpr,
             "dockerfile": OptLocalPathExpr,
-            "build_args": SimpleSeq(StrExpr),
-            "env": SimpleMapping(StrExpr),
-            "volumes": SimpleSeq(OptStrExpr),
+            "build_args": ExprOrSeq(StrExpr, type2str),
+            "env": ExprOrMapping(StrExpr, type2str),
+            "volumes": ExprOrSeq(OptStrExpr, type2str),
+            "build_preset": OptStrExpr,
         },
         ast.Image,
     )
@@ -495,6 +557,38 @@ FlowLoader.add_path_resolver("flow:images", [(dict, "images")])  # type: ignore
 FlowLoader.add_constructor("flow:images", parse_images)  # type: ignore
 
 
+def parse_needs_key(ctor: BaseConstructor, node: yaml.Node) -> IdExpr:
+    tmp = ctor.construct_scalar(node)  # type: ignore[no-untyped-call]
+    return IdExpr(mark2pos(node.start_mark), mark2pos(node.end_mark), tmp)
+
+
+FlowLoader.add_path_resolver(  # type: ignore
+    "flow:task_needs_key",
+    [(dict, "tasks"), (list, None), (dict, "needs"), (dict, True)],
+)
+FlowLoader.add_constructor("flow:task_needs_key", parse_needs_key)  # type: ignore
+
+
+def parse_needs(
+    ctor: BaseConstructor, node: yaml.Node
+) -> Optional[Mapping[IdExpr, NeedsLevel]]:
+    if isinstance(node, yaml.MappingNode):
+
+        def _factory(_start: Pos, _end: Pos, value: Any) -> NeedsLevel:
+            return NeedsLevel(value)
+
+        return SimpleMapping(_factory).construct(ctor, node)  # type: ignore
+    needs = SimpleSeq(IdExpr).construct(ctor, node)
+    return {task_id: NeedsLevel.COMPLETED for task_id in needs}  # type: ignore
+
+
+FlowLoader.add_path_resolver(  # type: ignore
+    "flow:task_needs",
+    [(dict, "tasks"), (list, None), (dict, "needs")],
+)
+FlowLoader.add_constructor("flow:task_needs", parse_needs)  # type: ignore
+
+
 def parse_exc_inc(
     ctor: BaseConstructor, node: yaml.MappingNode
 ) -> Sequence[Mapping[str, StrExpr]]:
@@ -587,9 +681,9 @@ EXEC_UNIT = {
     "bash": OptBashExpr,
     "python": OptPythonExpr,
     "workdir": OptRemotePathExpr,
-    "env": SimpleMapping(StrExpr),
-    "volumes": SimpleSeq(OptStrExpr),
-    "tags": SimpleSeq(StrExpr),
+    "env": ExprOrMapping(StrExpr, type2str),
+    "volumes": ExprOrSeq(OptStrExpr, type2str),
+    "tags": ExprOrSeq(StrExpr, type2str),
     "life_span": OptTimeDeltaExpr,
     "http_port": OptIntExpr,
     "http_auth": OptBoolExpr,
@@ -600,7 +694,7 @@ EXEC_UNIT = {
 JOB = {
     "detach": OptBoolExpr,
     "browse": OptBoolExpr,
-    "port_forward": SimpleSeq(PortPairExpr),
+    "port_forward": ExprOrSeq(PortPairExpr, port_pair_item),
     "multi": SimpleOptBoolExpr,
     "params": None,
     **EXEC_UNIT,
@@ -613,6 +707,16 @@ JOB_ACTION_CALL = {
 }
 
 
+def check_extra(
+    node: yaml.Node, dct: Dict[str, Any], ret: Dict[str, Any], name: str
+) -> None:
+    diff = dct.keys() - ret.keys()
+    if diff:
+        raise ConstructorError(
+            f"{','.join(diff)} are not allowed in {name}", node.start_mark
+        )
+
+
 def select_shells(
     ctor: BaseConstructor, node: yaml.MappingNode, dct: Dict[str, Any]
 ) -> Dict[str, Any]:
@@ -637,10 +741,13 @@ def select_job_or_action(
     ctor: BaseConstructor, node: yaml.MappingNode, dct: Dict[str, Any]
 ) -> Dict[str, Any]:
     if "action" in dct:
-        return {k: v for k, v in dct.items() if k in JOB_ACTION_CALL}
+        ret = {k: v for k, v in dct.items() if k in JOB_ACTION_CALL}
+        check_extra(node, dct, ret, "action call")
     else:
-        dct2 = {k: v for k, v in dct.items() if k in JOB}
-        return select_shells(ctor, node, dct2)
+        tmp = {k: v for k, v in dct.items() if k in JOB}
+        check_extra(node, dct, tmp, "job")
+        ret = select_shells(ctor, node, tmp)
+    return ret
 
 
 JOB_OR_ACTION = {**JOB, **JOB_ACTION_CALL}
@@ -690,9 +797,9 @@ FlowLoader.add_path_resolver("flow:jobs", [(dict, "jobs")])  # type: ignore
 FlowLoader.add_constructor("flow:jobs", parse_jobs)  # type: ignore
 
 
-TASK = {
+TASK: Mapping[str, Any] = {
     "id": OptIdExpr,
-    "needs": SimpleSeq(IdExpr),
+    "needs": None,
     "strategy": None,
     "enable": EnableExpr,
     "cache": None,
@@ -702,7 +809,7 @@ TASK = {
 
 TASK_ACTION_CALL = {
     "id": OptIdExpr,
-    "needs": SimpleSeq(IdExpr),
+    "needs": None,
     "strategy": None,
     "enable": EnableExpr,
     "cache": None,
@@ -718,10 +825,13 @@ def select_task_or_action(
     ctor: BaseConstructor, node: yaml.MappingNode, dct: Dict[str, Any]
 ) -> Dict[str, Any]:
     if "action" in dct:
-        return {k: v for k, v in dct.items() if k in TASK_ACTION_CALL}
+        ret = {k: v for k, v in dct.items() if k in TASK_ACTION_CALL}
+        check_extra(node, dct, ret, "action call")
     else:
-        dct2 = {k: v for k, v in dct.items() if k in TASK}
-        return select_shells(ctor, node, dct2)
+        tmp = {k: v for k, v in dct.items() if k in TASK}
+        check_extra(node, dct, tmp, "task")
+        ret = select_shells(ctor, node, tmp)
+    return ret
 
 
 def find_task_type(
@@ -764,8 +874,8 @@ def parse_flow_defaults(ctor: FlowLoader, node: yaml.MappingNode) -> ast.FlowDef
             ctor,
             node,
             {
-                "tags": SimpleSeq(StrExpr),
-                "env": SimpleMapping(StrExpr),
+                "tags": ExprOrSeq(StrExpr, type2str),
+                "env": ExprOrMapping(StrExpr, type2str),
                 "workdir": OptRemotePathExpr,
                 "life_span": OptTimeDeltaExpr,
                 "preset": OptStrExpr,
@@ -778,8 +888,8 @@ def parse_flow_defaults(ctor: FlowLoader, node: yaml.MappingNode) -> ast.FlowDef
             ctor,
             node,
             {
-                "tags": SimpleSeq(StrExpr),
-                "env": SimpleMapping(StrExpr),
+                "tags": ExprOrSeq(StrExpr, type2str),
+                "env": ExprOrMapping(StrExpr, type2str),
                 "workdir": OptRemotePathExpr,
                 "life_span": OptTimeDeltaExpr,
                 "preset": OptStrExpr,
@@ -1131,7 +1241,7 @@ ACTION = {
 }
 
 
-def preprocess_action(
+def select_action(
     ctor: BaseConstructor, node: yaml.MappingNode, dct: Dict[str, Any]
 ) -> Dict[str, Any]:
     kind = dct.get("kind")
@@ -1150,6 +1260,7 @@ def preprocess_action(
             f"missing mandatory key 'kind'",
             node.start_mark,
         )
+    check_extra(node, dct, ret, f"{kind.value} action")
 
     if kind == ast.ActionKind.LIVE:
         if "job" in dct and dct["job"].params is not None:
@@ -1221,7 +1332,7 @@ def parse_action_main(ctor: BaseConstructor, node: yaml.MappingNode) -> ast.Base
         node,
         ACTION,
         ast.BaseAction,
-        preprocess=preprocess_action,
+        preprocess=select_action,
         find_res_type=find_action_type,
     )
     return ret
